" init "
from .version import __version__
from .aioheoscontroller import AioHeosController, AioHeosException, SOURCE_LIST
from .aioheosupnp import AioHeosUpnp
from .aioheosplayer import AioHeosPlayer
from .aioheosgroup import AioHeosGroup

# __all__ = ['AioHeos', 'AioHeosUpnp']



#!/usr/bin/env python3
"""Heos python lib."""

import asyncio
import json
import logging
from concurrent.futures import CancelledError

from . import aioheosgroup
from . import aioheosplayer
from . import aioheosupnp

_LOGGER = logging.getLogger(__name__)

HEOS_PORT = 1255

GET_PLAYERS = 'player/get_players'
GET_PLAYER_INFO = 'player/get_player_info'
GET_PLAY_STATE = 'player/get_play_state'
SET_PLAY_STATE = 'player/set_play_state'
GET_MUTE_STATE = 'player/get_mute'
SET_MUTE_STATE = 'player/set_mute'
GET_VOLUME = 'player/get_volume'
SET_VOLUME = 'player/set_volume'
GET_NOW_PLAYING_MEDIA = 'player/get_now_playing_media'
GET_QUEUE = 'player/get_queue'
CLEAR_QUEUE = 'player/clear_queue'
PLAY_NEXT = 'player/play_next'
PLAY_PREVIOUS = 'player/play_previous'
PLAY_QUEUE = 'player/play_queue'
TOGGLE_MUTE = 'player/toggle_mute'

GET_GROUPS = 'group/get_groups'
SET_GROUP = 'group/set_group'

BROWSE = 'browser/browse'

EVENT_PLAYER_VOLUME_CHANGED = 'event/player_volume_changed'
EVENT_PLAYER_STATE_CHANGED = 'event/player_state_changed'
EVENT_PLAYERS_CHANGED = 'event/players_changed'
EVENT_PLAYER_NOW_PLAYING_CHANGED = 'event/player_now_playing_changed'
EVENT_PLAYER_NOW_PLAYING_PROGRESS = 'event/player_now_playing_progress'
EVENT_PLAYER_QUEUE_CHANGED = 'event/player_queue_changed'

EVENT_USER_CHANGED = 'event/user_changed'
EVENT_SOURCES_CHANGED = 'event/sources_changed'
EVENT_GROUPS_CHANGED = 'event/groups_changed'
EVENT_GROUP_VOLUME_CHANGED = 'event/group_volume_changed'
EVENT_REPEAT_MODE_CHANGED = "event/repeat_mode_changed"
EVENT_SHUTTLE_MODE_CHANGED = "event/shuffle_mode_changed"

SYSTEM_PRETTIFY = 'system/prettify_json_response'
SYSTEM_REGISTER_FOR_EVENTS = 'system/register_for_change_events'
SYSTEM_SIGNIN = 'system/sign_in'
SYSTEM_SIGNOUT = 'system/sign_out'

BROWSE_MUSIC_SOURCES = 'browse/get_music_sources'
BROWSE_SEARCH = 'browse/search'
BROWSE_BROWSE = 'browse/browse'
BROWSE_SEARCH_CRITERIA = 'browse/get_search_criteria'
BROWSE_PLAY_STREAM = 'browse/play_stream'

SOURCE_LIST = {
    1: 'Pandora',
    2: 'Rhapsody',
    3: 'TuneIn',
    4: 'Spotify',
    5: 'Deezer',
    6: 'Napster',
    7: 'iHeartRadio',
    8: 'Sirius XM',
    9: 'Soundcloud',
    10: 'Tidal',
    11: 'Future service',
    13: 'Amazon Music',
    14: 'Future service',
    15: 'Moodmix',
    17: 'Future service',
    18: 'QQMusic'
}


class AioHeosException(Exception):
    """AioHeosException class."""

    # pylint: disable=super-init-not-called
    def __init__(self, message):
        self.message = message


class AioHeosController():
    """Asynchronous Heos class."""

    # ddpylint: disable=too-many-public-methods,too-many-instance-attributes
    def __init__(self,
                 loop,
                 host=None,
                 username=None,
                 password=None,
                 new_device_callback=None):
        self._host = host
        self._loop = loop
        self._username = username
        self._password = password
        self._need_login = bool(self._username)
        self._new_device_callback = new_device_callback
        self._players = None
        self._groups = None

        self._player_id = None
        self._upnp = None
        self._reader = None
        self._writer = None
        self._subscribtion_task = None

        self._favourites = []
        self._favourites_sid = None
        self._music_sources = {}

    async def ensure_player(self):
        """Ensure player."""
        # timeout after 10 sec
        self.request_players()
        for _ in range(0, 20):
            if self.player_id:
                return
            await asyncio.sleep(0.5)

    async def ensure_group(self):
        """Ensure group."""
        # timeout after 10 sec
        self.request_groups()
        for _ in range(0, 20):
            if self._groups is not None:
                return
            await asyncio.sleep(0.5)

    async def ensure_login(self):
        """Ensure login."""
        # timeout after 20 sec
        self.login()
        for _ in range(0, 20):
            if not self._need_login:
                return
            await asyncio.sleep(0.5)

    async def ensure_favourites_loaded(self):
        """Ensure favourites loaded."""
        # timeout after 20 sec
        for _ in range(0, 20):
            if self._favourites:
                return
            await asyncio.sleep(0.5)

    @staticmethod
    def _url_to_addr(url):
        import re
        addr = re.search('https?://([^:/]+)[:/].*$', url)
        if addr:
            return addr.group(1)
        return None

    async def connect(self, host=None, port=HEOS_PORT, callback=None):
        """Connect to device."""
        if host:
            self._host = host
        elif not self._host:
            # discover
            if not self._upnp:
                self._upnp = aioheosupnp.AioHeosUpnp(loop=self._loop)
            url = await self._upnp.discover()
            self._host = self._url_to_addr(url)

        # connect
        _LOGGER.debug('[I] Connecting to %s:%s', self._host, port)
        await self._connect(self._host, port)

        # please, do not prettify json
        self.register_pretty_json(False)
        # and get events
        self.register_for_change_events()

        # setup subscription loop
        if not self._subscribtion_task:
            self._subscribtion_task = self._loop.create_task(
                self._async_subscribe(callback))

        # request for players
        await self.ensure_player()
        await self.ensure_group()
        if self._need_login:
            await self.ensure_login()
            self.request_music_sources()

    async def _connect(self, host, port=HEOS_PORT):
        """Connect."""
        while True:
            wait = 5
            try:
                # pylint: disable=line-too-long
                self._reader, self._writer = await asyncio.open_connection(
                    host, port, loop=self._loop)
                return
            except TimeoutError:
                _LOGGER.warning('[W] Connection timed out'
                                ', will try %s:%s again in %d seconds ...',
                                host, port, wait)
            except ConnectionRefusedError:
                wait = 30
                _LOGGER.warning('[W] Connection refused'
                                ', will try %s:%s again in %d seconds ...',
                                host, port, wait)
            except Exception as exc:  # pylint: disable=broad-except
                _LOGGER.error('[E] %s', exc)

            await asyncio.sleep(wait)

    def send_command(self, command, message=None):
        """Send command."""
        msg = 'heos://' + command
        if message:
            if not message.get('pid'):
                message['pid'] = self.player_id
            msg += '?' + '&'.join("{}={}".format(key, val)
                                  for (key, val) in message.items())
        msg += '\r\n'
        _LOGGER.debug(msg)
        self._writer.write(msg.encode('ascii'))

    @staticmethod
    def _parse_message(message):
        """Parse message."""
        result = {}
        if message:
            for elem in message.split('&'):
                parts = elem.split('=')
                if len(parts) == 2:
                    result[parts[0]] = parts[1]
                elif len(parts) == 1:
                    result[parts[0]] = True
                else:
                    _LOGGER.warning('[W] No parts found in %s', message)

        return result

    def _handle_error(self, message):
        eid = message['eid']
        if eid == '2':
            pid = message['pid']
            player = self.get_player(pid)
            player.play_state = None
            raise AioHeosException('Player {} is offline'.format(pid))
        else:
            raise AioHeosException(message)

    def _dispatcher(self, command, message, payload):
        """Call parser functions."""
        _LOGGER.debug('DISPATCHER')
        _LOGGER.debug('[D] %s %s %s', command, message, payload)
        callbacks = {
            GET_PLAYERS:
            self._parse_players,
            GET_GROUPS:
            self._parse_groups,
            SET_GROUP:
            self._parse_set_group,
            GET_PLAY_STATE:
            self._parse_play_state,
            SET_PLAY_STATE:
            self._parse_play_state,
            GET_MUTE_STATE:
            self._parse_mute_state,
            SET_MUTE_STATE:
            self._parse_mute_state,
            GET_VOLUME:
            self._parse_volume,
            SET_VOLUME:
            self._parse_volume,
            GET_NOW_PLAYING_MEDIA:
            self._parse_now_playing_media,
            EVENT_PLAYER_VOLUME_CHANGED:
            self._parse_player_volume_changed,
            EVENT_GROUP_VOLUME_CHANGED:
            self._parse_group_volume_changed,
            EVENT_PLAYER_STATE_CHANGED:
            self._parse_player_state_changed,
            EVENT_PLAYERS_CHANGED:
            self._parse_players_changed,
            EVENT_PLAYER_NOW_PLAYING_CHANGED:
            self._parse_player_now_playing_changed,
            EVENT_PLAYER_NOW_PLAYING_PROGRESS:
            self._parse_player_now_playing_progress,
            EVENT_GROUPS_CHANGED:
            self._parse_groups_changed,
            SYSTEM_SIGNIN:
            self._parse_system_signin,
            BROWSE_MUSIC_SOURCES:
            self._parse_browse_music_source,
            BROWSE_BROWSE:
            self._parse_browse_browse,
        }
        commands_ignored = (SYSTEM_PRETTIFY, SYSTEM_REGISTER_FOR_EVENTS,
                            EVENT_PLAYER_QUEUE_CHANGED, EVENT_SOURCES_CHANGED,
                            EVENT_USER_CHANGED, EVENT_SHUTTLE_MODE_CHANGED,
                            EVENT_REPEAT_MODE_CHANGED)
        if command in callbacks:
            callbacks[command](payload, message)
        elif command in commands_ignored:
            _LOGGER.debug('[D] command "%s" is ignored.', command)
        else:
            _LOGGER.debug('[D] command "%s" is not handled.', command)

    def _parse_command(self, data):
        " parse command "
        try:
            data_heos = data['heos']
            command = data_heos['command']
            message = {}
            if 'message' in data_heos:
                if data_heos['message'].startswith('command under process'):
                    return None
                message = self._parse_message(data_heos['message'])
            if 'result' in data_heos.keys() and data_heos['result'] == 'fail':
                self._handle_error(message)
            if 'payload' in data.keys():
                self._dispatcher(command, message, data['payload'])
            elif 'message' in data_heos.keys():
                self._dispatcher(command, message, None)
            elif 'command' in data_heos.keys():
                self._dispatcher(command, None, None)
            else:
                raise AioHeosException(
                    'No message or payload in reply. payload {}'.format(data))
        # pylint: disable=bare-except
        except AioHeosException as exc:
            raise exc
        except Exception:
            _LOGGER.exception("Unexpected error for msg '%s'", data)
            raise AioHeosException('Problem parsing command.')

        return None

    async def _callback_wrapper(self, callback):
        if callback:
            try:
                await callback()
            except Exception:    # pylint: disable=broad-except
                pass

    async def _async_subscribe(self, callback=None):
        """ event loop """
        # pylint: disable=too-many-branches,logging-too-many-args
        while True:
            if not self._reader:
                await asyncio.sleep(0.1)
                continue
            try:
                msg = await self._reader.readline()
            except TimeoutError:
                _LOGGER.warning(
                    '[W] Connection got timed out, try to reconnect...')
                await self._connect(self._host)
            except ConnectionResetError:
                _LOGGER.warning(
                    '[W] Peer reset our connection, try to reconnect...')
                await self._connect(self._host)
            except (GeneratorExit, CancelledError):
                _LOGGER.info('[I] Cancelling event loop...')
                return
            except Exception as exc:    # pylint: disable=broad-except
                _LOGGER.error('[E] Ignoring', exc)
            _LOGGER.debug(msg.decode())
            # simplejson doesnt need to decode from byte to ascii
            data = json.loads(msg.decode())
            _LOGGER.debug('DATA:')
            _LOGGER.debug(data)
            try:
                self._parse_command(data)
            except AioHeosException as exc:
                _LOGGER.error('[E]', exc)
                _LOGGER.debug('MSG', msg)
                _LOGGER.debug('MSG decoded', msg.decode())
                _LOGGER.debug('MSG json', data)
                continue
            if callback:
                _LOGGER.debug('TRIGGER CALLBACK')
                self._loop.create_task(self._callback_wrapper(callback))

    def new_device_callback(self, callback):
        """Callback when new device."""
        self._new_device_callback = callback

    def close(self):
        " close "
        _LOGGER.info('[I] Closing down...')
        if self._subscribtion_task:
            self._subscribtion_task.cancel()

    def register_for_change_events(self):
        " register for change events "
        self.send_command(SYSTEM_REGISTER_FOR_EVENTS, {'enable': 'on'})

    def register_pretty_json(self, enable=False):
        " register for pretty json "
        set_enable = 'off'
        if enable:
            set_enable = 'on'
        self.send_command(SYSTEM_PRETTIFY, {'enable': set_enable})

    def request_players(self):
        " get players "
        self.send_command(GET_PLAYERS)

    def login(self):
        " login "
        self.send_command(SYSTEM_SIGNIN, {
            'un': self._username,
            'pw': self._password
        })

    def _parse_players(self, payload, _message):
        _players_json = payload
        self._player_id = _players_json[0]['pid']
        if not self._players:
            self._players = []

        for player in _players_json:
            old_player = self.get_player(player['pid'])
            if not old_player:
                new_player = aioheosplayer.AioHeosPlayer(self, player)
                self._players.append(new_player)
                if self._new_device_callback:
                    self._new_device_callback(new_player)
            else:
                old_player.player_info = player

    def _parse_groups(self, payload, _message):
        _groups_json = payload

        if not self._groups:
            self._groups = []

        group_copy = self._groups
        for group in _groups_json:
            old_group = self.get_group(group['gid'])
            if not old_group:
                new_group = aioheosgroup.AioHeosGroup(self, group)
                self._groups.append(new_group)
                if self._new_device_callback:
                    self._new_device_callback(new_group)
            else:
                old_group.player_info = group
                group_copy.remove(group)

        for remove_group in group_copy:
            # Make group offline
            remove_group.play_state = None

    def _parse_set_group(self, _payload, _message):
        self.request_groups()

    def _parse_system_signin(self, _payload, _message):
        self._need_login = False

    def get_players(self):
        """ get players array """
        return self._players

    def get_groups(self):
        """ get groups array """
        return self._groups

    def get_player(self, pid):
        """ get player from array """
        for player in self._players:
            if player.player_id == pid:
                return player
        return None

    def get_group(self, pid):
        """Get group from array."""
        for group in self._groups:
            if group.player_id == pid:
                return group
        return None

    @property
    def player_id(self):
        " get player id "
        return self._player_id

    def request_player_info(self, pid):
        " request player info "
        self.send_command(GET_PLAYER_INFO, {'pid': pid})

    def request_play_state(self, pid):
        " request play state "
        self.send_command(GET_PLAY_STATE, {'pid': pid})

    def _parse_play_state(self, _payload, message):
        self.get_player(message['pid']).play_state = message['state']
        if self.get_group(message['pid']):
            self.get_group(message['pid']).play_state = message['state']

    def request_mute_state(self, pid):
        " request mute state "
        self.send_command(GET_MUTE_STATE, {'pid': pid})

    def _parse_mute_state(self, _payload, message):
        self.get_player(message['pid']).mute = message['state']
        if self.get_group(message['pid']):
            self.get_group(message['pid']).mute = message['state']

    def request_volume(self, pid):
        " request volume "
        self.send_command(GET_VOLUME, {'pid': pid})

    def set_volume(self, volume_level, pid):
        " set volume "
        volume = min(100, max(0, volume_level))
        self.send_command(SET_VOLUME, {'pid': pid, 'level': volume})

    def _parse_volume(self, _payload, message):
        self.get_player(message['pid']).volume = float(message['level'])
        if self.get_group(message['pid']):
            self.get_group(message['pid']).volume = float(message['level'])

    def _set_play_state(self, state, pid=None):
        " set play state "
        if state not in ('play', 'pause', 'stop'):
            AioHeosException('Not an accepted play state {}.'.format(state))

        self.send_command(SET_PLAY_STATE, {
            'pid': pid if pid else self.player_id,
            'state': state
        })

    def stop(self, pid=None):
        " stop player "
        self._set_play_state('stop', pid)

    def play(self, pid=None):
        " play "
        self._set_play_state('play', pid)

    def pause(self, pid=None):
        " pause "
        self._set_play_state('pause', pid)

    def request_now_playing_media(self, pid=None):
        " get playing media "
        self.send_command(GET_NOW_PLAYING_MEDIA,
                          {'pid': pid if pid else self.player_id})

    def _parse_now_playing_media(self, payload, message):
        player = self.get_player(message["pid"])
        player.reset_now_playing()

        player.media_artist = payload.get('artist')
        player.media_album = payload.get('album')
        player.media_title = payload.get('song')
        player.media_image_url = payload.get('image_url')
        player.media_id = payload.get('mid')

        if 'sid' in payload:
            player.sid = payload['sid']
            if self._music_sources:
                source_obj = self._music_sources.get(player.sid,
                                                     {'name': 'unknown'})
                _LOGGER.debug("[D] SOURCE %s", source_obj)
                player.source_name = source_obj['name']

        player.qid = payload.get('qid')

        _LOGGER.debug("[D] _parse_now_playing_media %s", vars(player))

    def get_favourites(self):
        """ get duration """
        return self._favourites

    def request_queue(self, pid):
        " request queue "
        self.send_command(GET_QUEUE, {'pid': pid})

    def clear_queue(self, pid):
        " clear queue "
        self.send_command(CLEAR_QUEUE, {'pid': pid})

    def request_play_next(self, pid):
        " play next "
        self.send_command(PLAY_NEXT, {'pid': pid})

    def _parse_play_next(self, payload, message):
        " parse play next "
        pass    # pylint: disable=unnecessary-pass

    def request_play_previous(self, pid):
        " play prev "
        self.send_command(PLAY_PREVIOUS, {'pid': pid})

    def play_queue(self, pid, qid):
        " play queue "
        self.send_command(PLAY_QUEUE, {'pid': pid, 'qid': qid})

    def play_stream(self, pid, sid, mid):
        " play_stream "
        self.send_command(BROWSE_PLAY_STREAM, {
            'pid': pid,
            'mid': mid,
            'sid': sid
        })

    def play_favourite(self, pid, mid):
        " play play_favourite "
        self.send_command(BROWSE_PLAY_STREAM, {
            'pid': pid,
            'mid': mid,
            'sid': self._favourites_sid
        })

    def request_groups(self):
        " get groups "
        self.send_command(GET_GROUPS)

    def set_group(self, leader_pid, member_pids):
        " set group "
        members = str(leader_pid)
        for member in member_pids:
            members = members + "," + str(member)
        self.send_command(SET_GROUP, {'pid': members})

    def toggle_mute(self, pid):
        " toggle mute "
        self.send_command(TOGGLE_MUTE, {'pid': pid})

    def set_mute(self, pid, mute):
        " set mute "
        self.send_command(SET_MUTE_STATE, {
            'pid': pid,
            'state': 'on' if mute else 'off'
        })

    def request_music_sources(self):
        " get music sources "
        self.send_command(BROWSE_MUSIC_SOURCES, {'range': '0,29'})

    def request_browse_source(self, sid):
        " browse source "
        self.send_command(BROWSE, {'sid': sid, 'range': '0,29'})

    def play_content(self, content, content_type='audio/mpeg'):
        """ play content """
        self._loop.create_task(self._upnp.play_content(content, content_type))
        # asyncio.wait([task])

    def _parse_player_volume_changed(self, _payload, message):
        player = self.get_player(message["pid"])
        player.mute = message['mute']
        player.volume = float(message['level'])

    def _parse_group_volume_changed(self, _payload, message):
        group = self.get_group(message["gid"])
        if group:
            group.mute = message['mute']
            group.volume = float(message['level'])

    def _parse_player_state_changed(self, _payload, message):
        player = self.get_player(message["pid"])
        player.play_state = message['state']
        group = self.get_group(message["pid"])
        if group:
            group.play_state = message['state']

    def _parse_groups_changed(self, _payload, _message):
        self.request_groups()

    def _parse_players_changed(self, _payload, _message):
        self.request_players()

    def _parse_player_now_playing_changed(self, _payload, _message):
        " event / now playing changed, request what changed. "
        self.request_now_playing_media()

    def _parse_player_now_playing_progress(self, _payload, message):
        player = self.get_player(message["pid"])
        player.current_position = int(message['cur_pos'])
        player.duration = int(message['duration'])

    def _parse_browse_music_source(self, payload, _message):
        _LOGGER.debug("[D] _parse_browse_music_source %s", payload)
        self._music_sources = {}

        for source in payload:
            self._music_sources[source['sid']] = source
            _LOGGER.debug('[D] Source: %s', source)
            if source['name'] == 'TuneIn':
                _LOGGER.debug('[D] TuneIn %s', source)
            elif source['name'] == 'Favorites':
                _LOGGER.debug('[D] Favorites %s', source)
                self._favourites_sid = source['sid']
                self.send_command(BROWSE_BROWSE, {"sid": source['sid']})

    def _parse_browse_browse(self, payload, message):
        _LOGGER.debug('[D] BROWSE_BROWSE fav sid <%s>, input sid <%s>',
                      self._favourites_sid, message['sid'])
        if str(message['sid']) == str(self._favourites_sid):
            _LOGGER.debug("[D] Favorites: %s", payload)
            self._favourites = payload

    def get_music_sources(self):
        """Get music source."""
        source_names = []
        for source in self._music_sources.values():
            source_names.append(source['name'])
        for source in self._favourites:
            source_names.append("Favorites__" + source['name'])

        return source_names



#!/usr/bin/env python3
""" Heos python lib

TODO:
    * validate soapaction replies
    * unit tests
    # remove httlib2?

"""

import asyncio
import logging
import socket
from time import gmtime, strftime

import aiohttp
import lxml.etree

SSDP_HOST = '239.255.255.250'
SSDP_PORT = 1900

# DENON_DEVICE = 'urn:schemas-denon-com:device:ACT-Denon:1'
DENON_DEVICE = 'urn:schemas-denon-com:device:AiosDevice:1'    # for dlna
MEDIA_DEVICE = 'urn:schemas-upnp-org:device:MediaRenderer:1'
AVTRANSPORT_SERVICE = 'urn:schemas-upnp-org:service:AVTransport:1'

_LOGGER = logging.getLogger(__name__)


def _get_ipaddress():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.connect(("google.com", 80))
    ipaddress = sock.getsockname()[0]
    sock.close()
    return ipaddress


class HttpException(Exception):
    """HttpException class."""

    # pylint: disable=super-init-not-called
    def __init__(self, message):
        self.message = message


class Http():
    """HTTP."""

    def __init__(self, loop):    # pylint: disable=redefined-outer-name
        self._loop = loop
        self._headers = {}

    def add_header(self, key, value):
        " add header "
        self._headers[key] = value

    def _add_user_agent_header(self):
        self.add_header('user_agent', 'Mega client v0.0.1')

    def get_headers(self):
        " get header "
        headers = ""
        for header, value in self._headers.items():
            headers += "{}: {}\r\n".format(header, value)
        return headers

    @staticmethod
    def _parse_uri(uri):
        import re

        match = re.search('https?://([^:/]+)(:([0-9]+))?(.*)$', uri)
        host = match.group(1)
        port = match.group(3)
        if not port:
            port = 80
        path = match.group(4)
        return (host, port, path)

    async def request(self, uri, method, data=None, headers=None):
        """ request """
        if headers is None:
            headers = {}
        host, port, path = Http._parse_uri(uri)

        method = "{method} {path} HTTP/1.0\r\n".format(
            method=method, path=path)
        self._add_user_agent_header()
        self._headers.update(headers)
        request = method.encode() + self.get_headers().encode() + data

        reader, writer = await asyncio.open_connection(
            host, port, loop=self._loop)
        writer.write(request.encode())

        reply = await reader.read()
        writer.close()
        return reply


class HttpResponse():
    """ HttpResponse """

    def __init__(self, status):
        self._headers = {}
        self._status = status
        self._add_server_header()
        self._add_date_header()

    def _add_server_header(self):
        self.add_header('Server', 'Mega Server v0.0.1')

    def _add_date_header(self):
        date = strftime("%a, %d %b %Y %H:%M:%S %Z", gmtime())
        self.add_header('Date', date)

    def add_header(self, key, value):
        " add header "
        self._headers[key] = value

    def get_headers(self):
        " get header "
        headers = ""
        for header, value in self._headers.items():
            headers += "{}: {}\r\n".format(header, value)
        return headers

    def get_status(self):
        " get status "
        return "HTTP/1.1 {status} OK\r\n".format(status=self._status)


class UpnpException(Exception):
    " UpnpException class "

    # pylint: disable=super-init-not-called
    def __init__(self, message):
        self.message = message


class Upnp():
    " Upnp class "

    # pylint: disable=redefined-outer-name
    def __init__(self, loop, ssdp_host=SSDP_HOST, ssdp_port=SSDP_PORT):
        self._loop = loop
        self._ssdp_host = ssdp_host
        self._ssdp_port = ssdp_port
        self._url = None
        # Backward compat
        if hasattr(asyncio, 'ensure_future'):
            self._asyncio_ensure_future = getattr(asyncio, 'ensure_future')
        else:
            self._asyncio_ensure_future = getattr(asyncio, 'async')

    class DiscoverProtocol:
        """ Discovery Protocol """

        def __init__(self, upnp, future, search_target):
            self._upnp = upnp
            self._future = future
            self._search_target = search_target
            self._transport = None

        def connection_made(self, transport):
            """ Protocol connection made """
            _LOGGER.debug('Connection made')
            self._transport = transport
            sock = self._transport.get_extra_info('socket')
            # sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.settimeout(2)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
            sock.bind(('', self._upnp.ssdp_port))

            tmpl = (
                'M-SEARCH * HTTP/1.1',
                'Host: ' + self._upnp.ssdp_host + ':' +
                str(self._upnp.ssdp_port),
                'Man: "ssdp:discover"',
                'ST: {}'.format(self._search_target),
                # 'ST: ssdp:all',
                'MX: 3',
                '',
                '')

            msg = "\r\n".join(tmpl).encode('ascii')
            self._transport.sendto(
                msg, (self._upnp.ssdp_host, self._upnp.ssdp_port))

        def datagram_received(self, data, _):
            """ datagram received """
            content = data.decode().rsplit('\r\n')

            # sock = self._transport.get_extra_info('socket')
            # if content[0].startswith('NOTIFY'):
            if content[0] == 'HTTP/1.1 200 OK':
                # pylint: disable=line-too-long
                reply = {
                    k.lower(): v
                    for k, v in
                    [item.split(': ') for item in content if ": " in item]
                }
                _LOGGER.debug(reply['st'])
                if reply['st'] == self._search_target:
                    url = reply['location']
                    self._future.set_result(url)
                    self._transport.close()

        def error_received(self, exc):    # pylint: disable=no-self-use
            """ error received """
            _LOGGER.error('[E] Error received: ', exc)

        def connection_lost(self, _):
            """ connection lost """
            _LOGGER.warning("[I] Connection lost.")
            self._transport.close()

    async def discover(self, search_target, _addr=None):
        " search "
        future = asyncio.Future()
        self._asyncio_ensure_future(
            self._loop.create_datagram_endpoint(
                lambda: Upnp.DiscoverProtocol(
                    self, future, search_target
                ),
                family=socket.AF_INET,
                proto=socket.IPPROTO_UDP))
        await future
        self._url = future.result()
        return self._url

    async def discover_mediarenderer(self, addr=None):
        " search media renderer "
        await self.discover(MEDIA_DEVICE, addr)

    # b'POST /AVTransport/control HTTP/1.1
    #   Host: XXXX:YYYY
    #   Content-Length: 560
    #   soapaction: "urn:schemas-upnp-org:service:AVTransport:1#SetAVTransportURI"
    #   accept-encoding: gzip, deflate
    #   user-agent: Python-httplib2/0.9.2 (gzip)
    #   content-type: text/xml; charset="utf-8"
    #   '
    async def _soapaction(self, service, action, url=None, body=None):
        " soap action "
        if not url:
            url = self._url

        headers = {
            'Content-Type': 'text/xml; charset="utf-8"',
            'SOAPAction': '"{}#{}"'.format(service, action)
        }

        content = ''
        with aiohttp.ClientSession(loop=self._loop) as session:
            response = await session.post(url, data=body, headers=headers)
            if response.status == 200:
                content = await response.read()
            await response.release()

        _LOGGER.debug(content)
        return content

    async def query_renderer(self, service, url=None):
        " query renderer "
        if not url:
            url = self._url

        # query renderer
        with aiohttp.ClientSession(loop=self._loop) as session:
            response = await session.get(url)
            if response.status == 200:
                content = await response.read()
            await response.release()

        # parse
        xml = lxml.etree.fromstring(content)    # pylint: disable=no-member
        # print(lxml.etree.tostring(xml, pretty_print=True).decode())
        xpath_search = '//n:service[n:serviceType="{}"]/n:controlURL/text()'.format(
            service)
        path = xml.xpath(
            xpath_search, namespaces={'n': 'urn:schemas-upnp-org:device-1-0'})
        try:
            return path[0]
        except TypeError:
            raise UpnpException('Cant find renderer')

    async def set_avtransport_uri(self, uri, url=None):
        " load url "
        service = AVTRANSPORT_SERVICE
        action = "SetAVTransportURI"
        # pylint: disable=line-too-long,bad-continuation
        body = (
            '<?xml version="1.0" encoding="utf-8"?>'
            '<s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">'
            '<s:Body>'
            '<u:SetAVTransportURI xmlns:u="{service}">'
            '<InstanceID>0</InstanceID>'
            '<CurrentURI>{uri}</CurrentURI>'
            '<CurrentURIMetaData></CurrentURIMetaData>'
            '</u:SetAVTransportURI>'
            '</s:Body>'
            '</s:Envelope>').format(
                service=service, uri=uri)
        response_xml = await self._soapaction(service, action, url, body)
        _LOGGER.debug(response_xml)

    async def set_play(self, url=None):
        " play "
        service = AVTRANSPORT_SERVICE
        # pylint: disable=line-too-long,bad-continuation
        body = (
            '<?xml version="1.0" encoding="utf-8"?>'
            '<s:Envelope s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" xmlns:s="http://schemas.xmlsoap.org/soap/envelope/">'
            '<s:Body>'
            '<u:Play xmlns:u="{}">'
            '<InstanceID>0</InstanceID>'
            '<Speed>1</Speed>'
            '</u:Play>'
            '</s:Body>'
            '</s:Envelope>').format(service)
        response_xml = await self._soapaction(service, "Play", url, body)
        _LOGGER.debug(response_xml)

    @property
    def ssdp_host(self):
        """ return ssdp host """
        return self._ssdp_host

    @property
    def ssdp_port(self):
        """ return ssdp port """
        return self._ssdp_port


class PlayContentServer(asyncio.Protocol):
    """ Play Content Server """

    def __init__(self, content, content_type):
        self._content = content
        self._content_type = content_type
        self._transport = None

    def connection_made(self, transport):
        self._transport = transport

    def data_received(self, data):
        # dont care of request, sent data anyway...
        _LOGGER.debug(data)

        response = HttpResponse(200)
        # response.add_header('Content-Length', content.getbuffer().nbytes)
        response.add_header('Content-Length', len(self._content))
        response.add_header('Content-Type', self._content_type)
        headers = response.get_status() + response.get_headers() + '\r\n'
        self._transport.write(headers.encode() + self._content)
        self._transport.close()


class AioHeosUpnp():
    " Heos version of Upnp "

    def __init__(self, loop):
        self._loop = loop
        self._upnp = None
        self._url = None
        self._path = None
        self._renderer_uri = None

    async def discover(self):
        " discover "
        if not self._upnp:
            self._upnp = Upnp(loop=self._loop)
        self._url = await self._upnp.discover(DENON_DEVICE)
        return self._url

    async def query_renderer(self):
        " query renderer "
        if not self._url or not self._upnp:
            return
        self._path = await self._upnp.query_renderer(
            AVTRANSPORT_SERVICE, self._url)
        self._renderer_uri = self._url + self._path

    async def _play_uri(self, uri):
        " play an url "
        if not self._url:
            await self.discover()
        if not self._renderer_uri:
            await self.query_renderer()
        await self._upnp.set_avtransport_uri(uri, self._renderer_uri)
        await self._upnp.set_play(self._renderer_uri)

    async def play_content(self, content, content_type='audio/mpeg', port=0):
        " play "
        address = _get_ipaddress()

        # create a listening port
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('', port))
        sock.listen(1)
        port = sock.getsockname()[1]

        uri = 'http://{}:{}/dummy.mp3'.format(address, port)

        # http server
        http_server = self._loop.create_server(
            lambda: PlayContentServer(content, content_type),
            sock=sock
        )

        # play request
        play_uri = self._loop.create_task(self._play_uri(uri))

        await asyncio.wait([http_server, play_uri])


async def main(aioloop):    # pylint: disable=redefined-outer-name
    " main "
    from pprint import pprint

    upnp = AioHeosUpnp(loop=aioloop)
    url = await upnp.discover()
    pprint(url)

    print('Query renderer')
    await upnp.query_renderer()

    content = b''
    with open('hello.mp3', mode='rb') as fhello:
        content = fhello.read()
    content_type = 'audio/mpeg'
    http_port = 8888

    print('Play content')
    await upnp.play_content(content, content_type, http_port)
    print('Play done')


if __name__ == "__main__":
    # pylint: disable=invalid-name
    aioloop = asyncio.get_event_loop()
    main(aioloop)
    aioloop.close()



#!/usr/bin/env python3
" Heos Group "

import logging
from . import aioheosplayer

_LOGGER = logging.getLogger(__name__)


class AioHeosGroup(aioheosplayer.AioHeosPlayer):
    " Asynchronous Heos Group class "

    def __init__(self, controller, group_json):
        group_json["pid"] = group_json["gid"]
        super().__init__(controller, group_json)
        _LOGGER.debug("[D] Creating group object %s for controller pid %s",
                      self._player_id, self._controller._player_id)

    def recreate_group(self):
        " Recreate group "
        member_ids = []
        for player in self._player_info["players"]:
            if str(player["pid"]) != str(self.player_id):
                member_ids.append(player["pid"])
        self._controller.set_group(self.player_id, member_ids)



#!/usr/bin/env python3
" Heos Player "

from pytz import UTC
from datetime import datetime

import logging

_LOGGER = logging.getLogger(__name__)


class AioHeosPlayer():
    " Asynchronous Heos Player class "

    def __init__(self, controller, player_json):
        self._player_info = player_json
        self._player_id = str(player_json['pid'])
        self._controller = controller
        self._sid = None
        self._source_name = None
        self._qid = None
        self._online = False
        self._play_state = None
        self._mute_state = None
        self._volume_level = 0
        self._current_position = 0
        self._current_position_updated_at = None
        self._duration = 0
        self._media_artist = None
        self._media_album = None
        self._media_title = None
        self._media_image_url = None
        self._media_id = None
        self._callback = None
        _LOGGER.debug("[D] Creating player object %s for controller pid %s",
                      self._player_id, self._controller._player_id)

    def __lt__(self, other):
        return self._player_id < other.player_id

    @property
    def state_change_callback(self):
        " get state_change_callback "
        return self._callback

    @state_change_callback.setter
    def state_change_callback(self, callback):
        self._callback = callback

    @property
    def player_id(self):
        " get player id "
        return self._player_id

    @property
    def name(self):
        " get player name "
        return self._player_info['name']

    @property
    def ip_address(self):
        " get player name "
        return self._player_info['ip']

    @property
    def volume(self):
        " get volume "
        return self._volume_level

    @volume.setter
    def volume(self, value):
        self._volume_level = value
        self._online = True
        self.notify_listeners()

    @property
    def current_position_updated_at(self):
        " get current_position_updated_at "
        return self._current_position_updated_at

    @property
    def duration(self):
        " get dureation "
        return self._duration

    @duration.setter
    def duration(self, duration):
        self._duration = duration
        self.notify_listeners()

    @property
    def current_position(self):
        " get current_position "
        return self._current_position

    @current_position.setter
    def current_position(self, current_position):
        self._current_position = current_position
        self._current_position_updated_at = datetime.now(UTC)
        self.notify_listeners()

    @property
    def mute(self):
        " get mute "
        return self._mute_state

    @mute.setter
    def mute(self, value):
        self._mute_state = value
        self.notify_listeners()

    @property
    def play_state(self):
        " get play state "
        return self._play_state

    @play_state.setter
    def play_state(self, value):
        self._play_state = value
        self._online = bool(value)
        self.notify_listeners()

    @property
    def media_artist(self):
        " get media artist "
        return self._media_artist

    @media_artist.setter
    def media_artist(self, value):
        self._media_artist = value
        self.notify_listeners()

    @property
    def media_album(self):
        " get media album "
        return self._media_album

    @media_album.setter
    def media_album(self, value):
        self._media_album = value
        self.notify_listeners()

    @property
    def media_title(self):
        " get media title "
        return self._media_title

    @media_title.setter
    def media_title(self, value):
        self._media_title = value
        self.notify_listeners()

    @property
    def media_image_url(self):
        " get media image url "
        return self._media_image_url

    @media_image_url.setter
    def media_image_url(self, value):
        self._media_image_url = value
        self.notify_listeners()

    @property
    def media_id(self):
        " get media_id "
        return self._media_id

    @media_id.setter
    def media_id(self, value):
        self._media_id = value

    @property
    def online(self):
        """Return True if entity is available."""
        return self._online

    @property
    def player_info(self):
        """Player info"""
        return self._player_info

    @player_info.setter
    def player_info(self, info):
        self._player_info = info
        self.notify_listeners()

    def toggle_mute(self):
        " toggle mute "
        self._controller.toggle_mute(self.player_id)

    def set_mute(self, mute):
        " set mute "
        self._controller.set_mute(self.player_id, mute)

    def reset_now_playing(self):
        """Reset now playing"""
        self._media_artist = None
        self._media_album = None
        self._media_title = None
        self._media_image_url = None
        self._media_id = None

    def request_update(self):
        """Request update"""
        self._controller.request_play_state(self.player_id)
        self._controller.request_mute_state(self.player_id)
        self._controller.request_volume(self.player_id)
        self._controller.request_now_playing_media(self.player_id)

    def volume_level_up(self, step=10):
        " volume level up "
        self.set_volume(self._volume_level + step)

    def volume_level_down(self, step=10):
        " volume level down "
        self.set_volume(self._volume_level - step)

    def stop(self):
        " stop player "
        self._controller.stop(self.player_id)

    def play(self):
        " play "
        self._controller.play(self.player_id)

    def pause(self):
        " pause "
        self._controller.pause(self.player_id)

    def play_next(self):
        " next "
        self._controller.request_play_next(self.player_id)

    def play_prev(self):
        " prev "
        self._controller.request_play_previous(self.player_id)

    def play_favorite(self, fav_mid):
        " Favorites "
        self._controller.play_favourite(self.player_id, fav_mid)

    def play_stream(self, sid, mid):
        " Favorites "
        self._controller.play_stream(self.player_id, sid, mid)

    def play_source(self, source):
        " Sourced "
        for src in self._controller._music_sources:
            if src['name'] == source:
                self._controller.play_stream(self.player_id, src['sid'], None)

    def set_volume(self, volume):
        """Set volume"""
        self._controller.set_volume(volume, self.player_id)

    def source_list(self):
        """Source list"""
        return self._controller.get_music_sources()

    def favourites_list(self):
        """Favourites list"""
        return self._controller.get_favourites()

    def create_group(self, devices):
        """Create group"""
        slave_ids = [slave for slave in devices if slave != self.player_id]
        self._controller.set_group(self.player_id, slave_ids)

    @property
    def sid(self):
        """Sid"""
        return self._sid

    @sid.setter
    def sid(self, value):
        self._sid = value

    @property
    def source_name(self):
        """Source name"""
        return self._source_name

    @source_name.setter
    def source_name(self, value):
        self._source_name = value

    @property
    def qid(self):
        """Qid"""
        return self._qid

    @qid.setter
    def qid(self, value):
        self._qid = value

    def notify_listeners(self):
        """Notify listeners"""
        if self._callback:
            self._callback()



from setuptools import setup

# http://stackoverflow.com/questions/6357361/alternative-to-execfile-in-python-3
exec(compile(open('aioheos/version.py', 'rb').read(), 'aioheos/version.py', 'exec'))

setup(name='aioheos',
      version=__version__,
      description='Denon HEOS',
      url='http://github.com/easink/aioheos',
      author='Andreas Rydbrink',
      author_email='andreas.rydbrink@gmail.com',
      license='MIT',
      packages=['aioheos'],
      long_description=open('README.md').read(),
      install_requires=[
          'lxml',
          'aiohttp',
      ],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      zip_safe=False)
